"""Contains the project versioning."""
from appcommander.verification.arg_verification import verify_args

from .arg_parser.arg_parser import parse_cli_args
from .arg_parser.process_args import process_args

__version__ = "0.0.17"
__version_info__ = tuple(int(i) for i in __version__.split(".") if i.isdigit())


# Parse command line interface arguments to determine what this script does.
args = parse_cli_args()
verify_args(
    args,
)
process_args(
    args,
)



"""Stores the flow logic of the script in a networkx graph."""

import networkx as nx
from typeguard import typechecked


# pylint: disable=R0903
class Screen_flow:
    """Contains a map from screen number and action number to next expected
    screens."""

    # pylint: disable=R0904
    @typechecked
    def __init__(
        self,
    ) -> None:
        """Creates the networkx graph with the screen nrs as nodes, and the
        action lists as edges."""
        self.G = nx.DiGraph()
        for nodename in list(range(0, 8 + 1)):
            self.G.add_node(nodename)

        # Set root CA as trusted, on phone.

        # Set permissions of DAVx5 app.

        # Launch DAVx5 with adb command to pre-load configuration

        # Screen 0: CLick LOGIN.
        self.G.add_edge(0, 1, actions=[0])

        # Screen 1: querying the Nextcloud server over tor.
        self.G.add_edge(1, 1, actions=[0])  # Not found.
        self.G.add_edge(1, 2, actions=[0])  # Not found.
        self.G.add_edge(1, 3, actions=[0])  # Found. cert not yet trusted.
        self.G.add_edge(1, 4, actions=[0])  # Found, cert trusted.
        self.G.add_edge(1, 6, actions=[0])  # Done, enter DAVx5 accountname.

        # TODO: Was not able to connect to tor. Restart orbot and try again.
        self.G.add_edge(2, 0, actions=[0])

        # Screen 3: If root CA is not trusted, install then restart flow.
        self.G.add_edge(3, 0, actions=[0])
        self.G.add_edge(3, 1, actions=[0])
        # Or click: "trust cert" in app and get additional prompt in screen 3.
        self.G.add_edge(3, 4, actions=[0])
        self.G.add_edge(3, 5, actions=[0])
        self.G.add_edge(3, 6, actions=[0])

        # Screen 4: Give name for certificate and press ok.
        self.G.add_edge(4, 6, actions=[0])

        # If OK was pressed before cert was set, set again with warning message
        # in screen 4, and press ok.
        self.G.add_edge(4, 5, actions=[0])

        # Screen 5: If OK was pressed before cert was set, set again with
        # warning message, and press ok.
        self.G.add_edge(5, 6, actions=[0])

        # Screen 6: set the name of the Nextcloud account in DAVx5.
        self.G.add_edge(6, 7, actions=[0])

        self.G.add_edge(7, 8, actions=[0])



"""Functions to assist a script file for the DAVx5 app.."""
from pathlib import Path
from typing import Dict, List

from typeguard import typechecked

from appcommander.run_bash_code import run_bash_command


@typechecked
def install_self_signed_root_ca_on_android(
    app_version_dir: str,
) -> None:
    """Verifies the/a self-signed root ca file exists in the root dir of this
    repository.

    Then calls the bash function that installs it. Then returns the
    control back to the script flow for DAVx5 installation.
    """
    bash_script_path = f"{app_version_dir}bash_scripts/export_root_ca2.sh"
    ca_local_filepath: str = "ca.crt"
    # Verify the file exists.
    for filepath in [bash_script_path, ca_local_filepath]:
        if not Path(filepath).is_file():
            raise Exception(f"Error, filepath:{filepath} does not exist.")

    # TODO: execute commands individually with intermediate Python
    # verification, instead of in Bash script.

    command = (
        f"bash -c 'source {bash_script_path} && install_self_signe"
        + f'd_root_ca_on_android "{ca_local_filepath}"\''
    )

    print(f"command={command}")
    run_bash_command(
        await_compilation=True, bash_command=command, verbose=False
    )


@typechecked
def get_calendar_names(
    sub_screen_dict: Dict,
) -> List[str]:
    """Returns the list of the calendar names that are to be synced."""
    calendar_names_list: List[str] = []
    calendar_list = sub_screen_dict["node"]["node"]
    for index in list(range(0, len(calendar_list))):
        calendar_name = calendar_list[index]["node"][1]["@text"]
        calendar_names_list.append(calendar_name)
    return calendar_names_list



"""Verifies the given CLI arguments are valid in combination with each
other."""
import argparse
import os
from pathlib import Path
from typing import Dict, List, Tuple

from typeguard import typechecked

from appcommander.helper import make_path_if_not_exists
from appcommander.verification.verify_phone_connection import (
    assert_app_is_installed,
)


@typechecked
def sort_out_app_name_and_package_name(
    some_apk_name: str, app_name_mappings: Dict[str, str]
) -> Tuple[str, str]:
    """Returns app name and package name based on user input and supported apk
    packages."""

    if some_apk_name in app_name_mappings.keys():
        app_name = some_apk_name
        package_name = app_name_mappings[app_name]
    elif some_apk_name in app_name_mappings.values():
        for key in app_name_mappings.keys():
            if app_name_mappings[key] == some_apk_name:
                app_name = key
        package_name = some_apk_name
    return app_name, package_name


@typechecked
def verify_args(args: argparse.Namespace) -> None:
    """Performs the checks to verify the parser."""
    verify_app_name(args)
    verify_app_version(args)


@typechecked
def verify_app_name(args: argparse.Namespace) -> None:
    """Verifies the app name is specified and that its folder exists, unless
    the user only wants to store screenshots, in that case, the app name and
    version can be create automatically."""
    # Verify the app name is specified.
    if args.app_name is None:
        raise NameError("Error, app name is not specified.")

    app_path: str = (
        f'src/appcommander/{args.app_name.replace(".","_").replace(" ","_")}'
    )
    # User just wants to store screenshots and json, make dirs for user.
    if args.export_screen is not None:
        make_path_if_not_exists(app_path)
    else:  # User tries to use app name even though app name may not exist.
        if not os.path.exists(app_path):
            raise NotADirectoryError(
                f"Error, path:{app_path} does not exist. Please create it."
            )


@typechecked
def verify_app_version(args: argparse.Namespace) -> None:
    """Verifies the app version is specified and that its folder exists, unless
    the user only wants to store screenshots, in that case, the app version can
    be create automatically."""
    # Verify the app name is specified.
    if args.version is None:
        raise NameError("Error, app version is not specified.")

    version_path: str = (
        f'src/appcommander/{args.app_name.replace(".","_").replace(" ","_")}/'
        + f'V{args.version.replace(".","_").replace(" ","_")}'
    )
    # User just wants to store screenshots and json, make dirs for user.
    if args.export_screen is not None:
        # Create directories if not yet existent.
        make_path_if_not_exists(version_path)
    else:  # User tries to use app name even though app name may not exist.
        if not os.path.exists(version_path):
            raise NotADirectoryError(
                f"Error, path:{version_path} does not exist. "
                + " Please create it and add an accompanying script."
            )
        script_path = f"{version_path}/Screen_flow.py"
        if not os.path.exists(script_path):
            raise NotADirectoryError(
                f"Error, path:{script_path} does not exist. "
                + " Please create it and make it work."
            )


@typechecked
def verify_app_script(script_path: str) -> None:
    """Verifies the filepath for the input graph exists and contains a valid
    networkx graph."""

    # Assert graph file exists.
    if not Path(script_path).is_file():
        raise FileNotFoundError(f"Input Graph path was invalid:{script_path}")


def get_verified_apps_to_torify(
    app_name_mappings: Dict[str, str], torifying_aps_csv: str
) -> Dict[str, str]:
    """Converts the comma separated values (csv) of the app names that are to
    be torified, into a list of app names, and then verifies they are installed
    on the phone."""

    some_apps: List[str] = torifying_aps_csv.split(",")
    torryfying_apps: Dict[str, str] = {}
    for app_name in some_apps:
        app_name, package_name = sort_out_app_name_and_package_name(
            app_name, app_name_mappings=app_name_mappings
        )
        assert_app_is_installed(
            package_name=package_name,
        )
        torryfying_apps[app_name] = package_name

    return torryfying_apps



"""Stores the flow logic of the script in a networkx graph."""

from typeguard import typechecked

from appcommander.run_bash_code import run_bash_command


class App_input_data:
    """Stores the input data that is fed into the DAVx5 app."""

    # pylint: disable=R0903
    @typechecked
    def __init__(
        self,
        nextcloud_password: str,
        nextcloud_username: str,
        onion_url: str,
    ) -> None:
        """Creates the networkx graph with the screen nrs as nodes, and the
        action lists as edges."""
        self.nextcloud_password: str = nextcloud_password
        self.nextcloud_username: str = nextcloud_username
        self.onion_url: str = onion_url

    @typechecked
    def launch_app(
        self,
        package_name: str,
    ) -> None:
        """Launches DAVx5 with onion url of your Nextcloud server and your
        Nextcloud credentials."""

        print(f"Launching: {package_name}")
        # TODO: verify Nextcloud server is running on onion url.
        command = (
            "adb shell am start -a android.intent.action.VIEW -d caldavs://"
            + f"{self.nextcloud_username}:{self.nextcloud_password}@"
            + f"{self.onion_url}/remote.php"
            + f"/dav/principals/users/{self.nextcloud_username}"
        )

        print(f"command={command}")
        run_bash_command(
            await_compilation=True, bash_command=command, verbose=False
        )



"""Starts a script to control an app."""


from typing import Callable, Dict, List

from typeguard import typechecked
from uiautomator import AutomatorDevice

from appcommander.helper import export_screen_data_if_valid
from appcommander.Screen import Screen
from appcommander.Script import Script
from appcommander.verification.status_verification import can_proceed


@typechecked
def run_script(script: Script, dev: AutomatorDevice) -> None:
    """Runs the incoming script on the phone.

    Script folder structure: src/appcommander/app_name/version.py with
    app_name is something like: com_whatsapp_android (not: Whatsapp). It
    is derived from how your android dev calls the app, with the dots
    replaced by underscores. E.g. com.whatsapp.android or something like
    that.
    """

    # Open the app.
    # script.input_data.launch_app(app_name=script.app_name)
    script.input_data.launch_app(package_name=script.package_name)

    expected_screens: List[int] = list(
        map(lambda x: x.screen_nr, script.screens)
    )

    # First perorm a quick scope, without retry to find the desired screen.
    is_expected, screen_nr = can_proceed(
        dev=dev,
        expected_screennames=expected_screens,
        retry=False,
        script=script,
    )

    # If quickscope did not find desired screen, try again with retries and
    # waiting times per retry.
    if not is_expected:
        _, screen_nr = can_proceed(
            dev=dev,
            expected_screennames=expected_screens,
            retry=True,
            script=script,
        )
    script.past_screens.append(screen_nr)

    next_actions = "filler"
    retry: bool = False  # For the first screen, do a quick scope because it is
    # known already.
    while next_actions is not None and expected_screens:
        _, screen_nr = can_proceed(
            dev=dev,
            expected_screennames=expected_screens,
            retry=retry,
            script=script,
        )
        retry = True
        screen = script.script_graph.nodes[screen_nr]["Screen"]
        print(f"screen_nr={screen_nr}")

        # Export the data of the screens if they happen to be found in the
        # dev already.
        export_screen_data_if_valid(
            dev=dev,
            overwrite=script.overwrite,
            screens=[screen],
            script=script,
        )

        # Get next action
        next_actions = screen.get_next_actions(
            required_objects=screen.required_objects,
            optional_objects=screen.optional_objects,
            script=script,
        )

        # Perform next action.
        if next_actions is not None:

            # Compose the information needed for the actions.

            # Perform the actual action.
            action_output: Dict = perform_action(
                dev=dev,
                next_actions=next_actions,
                screen=screen,
                script=script,
            )
            expected_screens = action_output["expected_screens"]
            script.past_screens.append(screen_nr)

    print(f"Done with script:{script.app_name}")


@typechecked
def perform_action(
    dev: AutomatorDevice,
    next_actions: Callable,
    screen: Screen,
    script: Script,
) -> Dict:
    """Performs the first action list in the list of action lists."""
    action_output: Dict = next_actions(
        dev=dev,
        screen=screen,
        script=script,
    )
    if "expected_screens" not in action_output.keys():
        raise KeyError(
            "Error, the action output did not contain the expected screens."
        )
    return action_output



"""TODO Documentation."""
# pylint: disable=R0801
from typing import TYPE_CHECKING, Callable, Dict, List, Union

from typeguard import typechecked
from uiautomator import AutomatorDevice

from appcommander.Screen import Screen

if TYPE_CHECKING:
    from appcommander.Script import Script
else:
    Script = object


@typechecked
def screen_2() -> Screen:
    """Creates the settings for when DAVx5 is querying the Nextcloud
    server.."""
    max_retries = 5
    screen_nr = 2
    wait_time_sec = 1
    required_objects: List[Dict[str, str]] = [
        {
            "@package": "at.bitfire.davdroid",
        },
        {
            "@text": "Couldn't find CalDAV or CardDAV service.",
        },
    ]

    # pylint: disable=W0613
    @typechecked
    def get_next_actions(
        required_objects: List[Dict[str, str]],
        optional_objects: List[Dict[str, str]],
        script: Script,
    ) -> Union[Callable[[AutomatorDevice, Screen, Script], Dict], None]:
        """Looks at the required objects and optional objects and determines
        which actions to take next.
        An example of the next actions could be the following List:
        0. Select a textbox.
        1. Send some data to a textbox.
        2. Click on the/a "Next" button.

        Then the app goes to the next screen and waits a pre-determined
        amount, and optionally retries a pre-determined amount of attempts.
        """
        # In the start screen just press ok.
        return actions_0

    return Screen(
        get_next_actions=get_next_actions,
        is_start=True,
        max_retries=max_retries,
        screen_nr=screen_nr,
        wait_time_sec=wait_time_sec,
        required_objects=required_objects,
    )


# pylint: disable=W0613
@typechecked
def actions_0(dev: AutomatorDevice, screen: Screen, script: Script) -> Dict:
    """Performs the actions in option 1 in this screen.

    For this screen, it clicks the "OK" button in the "Connection
    request".
    """
    # TODO: Was not able to connect to tor. Restart orbot and try again.
    raise Exception(
        "Error, was not able to connect to TOR server, please:"
        + " Ensure your Nextcloud server is running and reachable over tor,"
        + " and ensure Orbot is torifying DAVx5, and connected to TOR."
    )



"""Functions to assist a script file for an arbitrary app."""
from typing import TYPE_CHECKING, Dict, List, Union, cast

from typeguard import typechecked

from appcommander.helper import dict_contains_other_dict

# pylint: disable=R0801
if TYPE_CHECKING:
    from appcommander.Screen import Screen
    from appcommander.Script import Script
else:
    Screen = object
    Script = object


@typechecked
def get_torified_item_index_dict(
    required_object: Dict[str, str],
    unpacked_screen_dict: Dict[str, Union[List, Dict, str]],
    parent_dict: Union[List, Dict[str, Union[List, Dict, str]]],
) -> Dict:  # -> Dict[str, str]:
    """Parses the screen dictionary and returns the index of the item of the
    app that you want to torify. The item is found by checking if it has a
    "node" key, which has a List of 3 elements as value (each in the form of a
    dict):
     - the app icon.
     - the app name
     - the checkbox to torify that app.
    The dict with the app name is used to identify the relevant item, then the
    dictionary of that item contains the @index value, which is the index of
    that app in the Orbot configuration/torify apps screen. This method returns
    the complete item dictionary including the 3 node children.

    It recursively looks through the screen dict until it has found the app."""

    if dict_contains_other_dict(required_object, unpacked_screen_dict):
        return cast(Dict, parent_dict)
    if "node" in unpacked_screen_dict.keys():
        if isinstance(unpacked_screen_dict["node"], Dict):
            if get_torified_item_index_dict(
                required_object=required_object,
                unpacked_screen_dict=unpacked_screen_dict["node"],
                parent_dict=unpacked_screen_dict,
            ):
                return get_torified_item_index_dict(
                    required_object=required_object,
                    unpacked_screen_dict=unpacked_screen_dict["node"],
                    parent_dict=unpacked_screen_dict,
                )
        if isinstance(unpacked_screen_dict["node"], List):

            # A item in the Orbot app consists of an icon, app_name and
            # checkbox. So if the length of this node List is three, and it
            # contains the app name in the dict of the first index of the list
            # elements, one wants the index of the actual dict that has this
            # node list, because that dict has the index of the item.
            if len(unpacked_screen_dict["node"]) == 3:
                if dict_contains_other_dict(
                    required_object, unpacked_screen_dict["node"][1]
                ):
                    return unpacked_screen_dict

            # Keep on searching deeper otherwise.
            for node_elem in unpacked_screen_dict["node"]:
                if get_torified_item_index_dict(
                    required_object=required_object,
                    unpacked_screen_dict=cast(Dict, node_elem),
                    parent_dict=unpacked_screen_dict["node"],
                ):
                    return get_torified_item_index_dict(
                        required_object=required_object,
                        unpacked_screen_dict=cast(Dict, node_elem),
                        parent_dict=unpacked_screen_dict["node"],
                    )
        if not isinstance(unpacked_screen_dict["node"], Dict | List):
            raise TypeError("Node value of unexpected type.")
    return {}


def orbot_torifying_app_is_checked(app_item: Dict) -> bool:
    """Returns True if the item is checked, False if it is unchecked."""
    if "@checkable" not in app_item["node"][2].keys():
        raise KeyError(
            "Error, the app checkbox should contain @checkable."
            + f"Though it did not:{app_item}"
        )
    if "@checkable" not in app_item["node"][2]["@checkable"] != "true":
        raise ValueError(
            "Error, the app checkbox should contain:"
            + f"@checkable: true Though it did not:{app_item}"
        )
    if app_item["node"][2]["@checked"] == "true":
        return True
    if app_item["node"][2]["@checked"] == "false":
        return False
    raise KeyError(
        "Error, valid @checked value was not found in app_item checkbox."
    )



"""Contains helper functions that are used throughout this repository."""
import json
import os
import time
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import xmltodict
from typeguard import typechecked
from uiautomator import AutomatorDevice

from appcommander.run_bash_code import run_bash_command

if TYPE_CHECKING:
    from appcommander.Screen import Screen
    from appcommander.Script import Script
else:
    Screen = object
    Script = object


@typechecked
def output_json(output_dir: str, filename: str, screen_dict: Dict) -> None:
    """Outputs a json to a filepath, ensures the output dir exists, overwrites
    json if it already exists."""

    make_path_if_not_exists(output_dir)

    output_path = f"{output_dir}{filename}"
    with open(output_path, "w", encoding="utf-8") as fp:
        json.dump(screen_dict, fp, indent=4, sort_keys=True)
        fp.close()

    # Verify the file exists.
    if not Path(output_path).is_file():
        raise Exception(f"Error, filepath:{output_path} was not created.")


@typechecked
def load_json_file_into_dict(
    json_filepath: str,
) -> Dict:
    """Loads a json file into dict from a filepath."""
    if not Path(json_filepath).is_file():
        raise Exception(f"Error, filepath does not exist:{json_filepath}")

    with open(json_filepath, encoding="utf-8") as json_file:
        the_dict = json.load(json_file)
        json_file.close()
    return the_dict


@typechecked
def make_path_if_not_exists(path: str) -> None:
    """Creates a filepath if it does not yet exist."""
    if not os.path.exists(path):
        os.makedirs(path)
    if not os.path.exists(path):
        raise Exception(f"Error, path:{path} did not exist after creation.")


@typechecked
def export_screen_data_if_valid(
    dev: AutomatorDevice,
    overwrite: bool,
    screens: List[Screen],
    script: Script,
) -> None:
    """Checks whether the required objects are in the actual screen, and if
    they are, it exports the data of the screen in json format and as a
    screenshot."""
    if dev is not None:
        for screen in screens:
            # Load and unpack the screen dict to get meaningful ui info.
            screen.screen_dict = get_screen_as_dict(
                dev=dev,
                unpack=True,
                screen_dict=screen.screen_dict,
                reload=False,
            )

            if is_expected_screen(
                dev=dev,
                unpacked_screen_dict=screen.screen_dict,
                retry=True,
                expected_screen=screen,
            ):
                export_screen_data(
                    dev=dev,
                    screen_dict=screen.screen_dict,
                    screen_nr=screen.screen_nr,
                    script=script,
                    overwrite=overwrite,
                    subdir="verified",
                )


@typechecked
def launch_app(app_name: str) -> None:
    """Launches app on phone."""

    # Launch the app on phone.
    command = f'adb shell monkey -p "{app_name}" 1 &>/dev/null'
    run_bash_command(
        await_compilation=True, bash_command=command, verbose=False
    )


# pylint: disable=R0913
@typechecked
def export_screen_data(
    dev: AutomatorDevice,
    screen_dict: Dict,
    screen_nr: int,
    script: Script,
    overwrite: bool = False,
    subdir: str = "unverified",
) -> None:
    """Writes a dict file to a .json file, and exports a screenshot.

    The overwrite bool determines whether a file will be overwritten or
    not if it  already exists. The subdir bool determines whether the
    screen data is verified to be expected, or not. If not, the screen
    data is placed in a subfolder named: subdir, to reduce the
    probability of the developer basing script actions on data belonging
    to the wrong screen.
    """
    output_dir = (
        (
            "src/appcommander/"
            + f"{script.package_name}"
            + f"/V{script.version}/{subdir}/"
        )
        .replace(".", "_")
        .replace(" ", "_")
    )
    output_name = f"{screen_nr}"

    for extension in [".json", ".png"]:
        output_path = f"{output_dir}{output_name}{extension}"
        if not Path(output_path).is_file() or overwrite:

            if extension == ".json":
                if screen_dict == {}:
                    screen_dict = get_screen_as_dict(
                        dev=dev,
                        unpack=True,
                        screen_dict=screen_dict,
                        reload=False,
                    )
                output_json(output_dir, f"{output_name}.json", screen_dict)
            if extension == ".png":
                # dev.takeScreenshot(output_path)
                dev.screenshot(output_path)

        # Verify the file exists.
        if not Path(output_path).is_file():
            raise Exception(f"Error, filepath:{output_path} was not created.")


@typechecked
def get_screen_as_dict(
    dev: AutomatorDevice,
    unpack: bool,
    screen_dict: Dict,
    reload: bool = False,
) -> Dict:
    """Loads the phone and shows the screen as a dict."""

    # Don't reload if the screen dict still exists, and no explicit
    # reload is asked.
    if screen_dict != {} and not reload:
        if unpack and "hierarchy" in screen_dict.keys():
            return screen_dict["hierarchy"]
        return screen_dict

    # Get the new screen data from the ui.
    if screen_dict == {} or reload:
        print("Loading screen data from phone for orientation.")
        ui_information: Dict = xmltodict.parse(dev.dump())

        # Unpack the screen dict to get a recursive dictionary structure.
        if unpack:
            ui_information = ui_information["hierarchy"]
    return ui_information


@typechecked
def is_expected_screen(
    dev: AutomatorDevice,
    expected_screen: Screen,
    retry: bool,
    unpacked_screen_dict: Dict,
    verbose: Optional[bool] = False,
) -> bool:
    """Custom verification per screen based on the optional and required
    objects in screen.

    Raise error if verification fails.
    """

    # Preliminary check to see if the required objects are in.
    if not required_objects_in_screen(
        expected_screen.required_objects, unpacked_screen_dict
    ):
        if not retry:
            return False
        # Retry and return True if the required objects were found.
        for _ in range(0, expected_screen.max_retries):

            # Reload the screen data again.
            unpacked_screen_dict = get_screen_as_dict(
                dev=dev,
                unpack=True,
                screen_dict={},
                reload=True,
            )
            print(
                f"Wait: {expected_screen.wait_time_sec} [s] on screen: "
                + f"{expected_screen.screen_nr}"
            )
            time.sleep(expected_screen.wait_time_sec)
            if required_objects_in_screen(
                required_objects=expected_screen.required_objects,
                unpacked_screen_dict=unpacked_screen_dict,
            ):
                return True
            if verbose:
                print(f"Not found:{expected_screen.required_objects}")
        if verbose:
            print(f"Not found:{expected_screen.required_objects}")
        return False
    # else:
    return True


@typechecked
def required_objects_in_screen(
    required_objects: List[Dict[str, Union[List, Dict, str]]],
    unpacked_screen_dict: Dict[str, Union[List, Dict, str]],
) -> bool:
    """Returns True if all required objects are found in the UI/screen.

    False otherwise.
    """
    for required_object in required_objects:
        if not required_object_in_screen(
            required_object=required_object,
            unpacked_screen_dict=unpacked_screen_dict,
        ):
            return False
    return True


@typechecked
def required_object_in_screen(
    required_object: Dict[str, Union[List, Dict, str]],
    unpacked_screen_dict: Dict[str, Union[List, Dict, str]],
) -> bool:
    """Returns True if all the keys and values in a dict are found within the
    same key or list of the xml of the ui."""
    if dict_contains_other_dict(required_object, unpacked_screen_dict):
        return True
    if "node" in unpacked_screen_dict.keys():
        if isinstance(unpacked_screen_dict["node"], Dict):
            if required_object_in_screen(
                required_object=required_object,
                unpacked_screen_dict=unpacked_screen_dict["node"],
            ):
                return True
        if isinstance(unpacked_screen_dict["node"], List):
            for node_elem in unpacked_screen_dict["node"]:
                if required_object_in_screen(
                    required_object=required_object,
                    unpacked_screen_dict=node_elem,
                ):
                    return True
        if not isinstance(unpacked_screen_dict["node"], Dict | List):
            raise TypeError("Node value of unexpected type.")
    return False


@typechecked
def dict_contains_other_dict(sub: Dict, main: Dict) -> bool:
    """Returns true if the sub dict is a subset of the main dict."""
    for sub_key, sub_val in sub.items():
        if sub_key not in main.keys():
            return False
        # An artifact like:
        # "@text": "VPN Mode \u200e\u200f\u200e\u200e\u200e\u200e\u200e\u200f
        # \u200e\u200f\u200f\u200f\u200e\u200e\u200e\u200e\u200e\u200e\u200f\
        # u200e\u200e\u200f\u200e\u200e\u200e\u200e\u200f\u200f\u200f\u200f\
        # u200f\u200e\u200e\u200f\u200f\u200e\u200e\u200e\u200f\u200e\u200e\
        # u200f\u200e\u200e\u200e\u200e\u200e\u200e\u200f\u200e\u200f\u200f\
        # u200f\u200e\u200e\u200e\u200e\u200e\u200e\u200f\u200e\u200f\u200e\
        # u200e\u200e\u200e\u200e\u200e\u200e\u200f\u200e\u200e\u200e\u200e\
        # u200f\u200e\u200e\u200e\u200f\u200f\u200f\u200f\u200f\u200e\u200e\
        # u200f\u200f\u200e\u200f\u200f\u200e\u200e\u200e\u200eON\u200e\u200f
        # \u200e\u200e\u200f\u200e",
        # may occur at random. Therefore, the "in" option is included.
        # TODO: determine why these artifacts may occur and/or remove them.
        if sub_val != main[sub_key] and sub_val not in main[sub_key]:
            return False
    return True



"""Script to run through Orbot app configuration.

Android names this app: org.torproject.android
"""


from typing import TYPE_CHECKING, Any, Dict, List, Union

from typeguard import typechecked

from appcommander.create_screens import create_screens, load_script_attribute

if TYPE_CHECKING:
    from appcommander.Screen import Screen
else:
    Screen = object


# pylint: disable=R0902
class Script:
    """Experiment manager.

    First prepares the environment for running the experiment, and then
    calls a private method that executes the experiment consisting of 4
    stages.
    """

    # pylint: disable=R0903
    # pylint: disable=R0913
    @typechecked
    def __init__(
        self,
        app_name: str,
        overwrite: bool,
        package_name: str,
        version: str,
        cli_input_data: Dict[str, Union[str, Dict[str, str]]],
    ) -> None:
        self.app_name: str = app_name
        self.overwrite: bool = overwrite
        self.package_name: str = package_name
        self.package_name_dir: str = self.package_name.replace(
            ".", "_"
        ).replace(" ", "_")
        # self.torifying_apps: Union[Dict[str, str], None] = torifying_apps
        self.version: str = version
        self.version_dir: str = self.version.replace(".", "_").replace(
            " ", "_"
        )

        self.app_version_mod_path: str = (
            f"appcommander.{self.package_name_dir}.V{self.version_dir}."
        )
        self.app_version_dir: str = (
            f"src/appcommander/{self.package_name_dir}/V{self.version_dir}/"
        )

        # Create placeholder for past screens.
        self.past_screens: List[int] = []

        # Generate the script screen flow as a graph and generate the screens.
        # self.script_graph = load_graph(self.app_version_mod_path)
        self.script_graph = load_script_attribute(
            app_version_mod_path=self.app_version_mod_path,
            filename="Screen_flow",
            obj_name="Screen_flow",
            attribute_name="G",
        )
        self.input_data = load_script_attribute(
            app_version_mod_path=self.app_version_mod_path,
            filename="App_input_data",
            obj_name="App_input_data",
        )
        self.input_data = fill_input_data(
            self.input_data,
            cli_input_data,
        )
        self.screens: List[Screen] = create_screens(self)


@typechecked
def fill_input_data(  # type:ignore[misc]
    input_data: Any,
    cli_input_data: Dict,
) -> Any:
    """Stores the CLI input data into the input_data object, with appropriate
    typing."""
    # The ** unpacks the dictionary into the object.
    # TODO: test what happens if the dictionary is under- or overcomplete.
    return input_data(**cli_input_data)



"""Entry point for this project, runs the project code based on the cli command
that invokes this script."""


from appcommander.verification.arg_verification import verify_args

from .arg_parser.arg_parser import parse_cli_args
from .arg_parser.process_args import process_args

# Parse command line interface arguments to determine what this script does.
args = parse_cli_args()
verify_args(
    args,
)
process_args(
    args,
)



"""Stores the flow logic of the script in a networkx graph."""

from typing import Dict

from typeguard import typechecked

from appcommander.run_bash_code import run_bash_command


class App_input_data:
    """Stores the input data that is fed into the DAVx5 app."""

    # pylint: disable=R0903
    @typechecked
    def __init__(
        self,
        torifying_apps: Dict[str, str],
    ) -> None:
        """Creates the networkx graph with the screen nrs as nodes, and the
        action lists as edges."""
        self.torifying_apps: Dict[str, str] = torifying_apps

    @typechecked
    def launch_app(
        self,
        package_name: str,
    ) -> None:
        """Launches the Orbot app."""
        print(f"Launching:{package_name}")

        # Launch the app on phone.
        command = f'adb shell monkey -p "{package_name}" 1 &>/dev/null'
        run_bash_command(
            await_compilation=True, bash_command=command, verbose=False
        )



"""The settings screen where apps are torified."""
# pylint: disable=R0801
import inspect
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Union

import networkx as nx
from typeguard import typechecked
from uiautomator import AutomatorDevice

from appcommander.helper import get_screen_as_dict
from appcommander.org_torproject_android.V16_6_3_RC_1.helper import (
    get_torified_item_index_dict,
    orbot_torifying_app_is_checked,
)
from appcommander.Screen import Screen
from appcommander.script_orientation import get_expected_screen_nrs

if TYPE_CHECKING:
    from appcommander.Script import Script
else:
    Script = object


@typechecked
def screen_6() -> Screen:
    """Creates the settings for a starting screen where Orbot is not yet
    started."""
    max_retries = 5
    screen_nr = 6
    wait_time_sec = 2
    required_objects: List[Dict[str, str]] = [
        # {
        # "@text": "Global " "(Auto)",
        # },
        {
            "@resource-id": "org.torproject.android:id/itemicon",
        },
        {
            "@resource-id": "org.torproject.android:id/itemtext",
        },
        {
            "@resource-id": "org.torproject.android:id/itemcheck",
        },
    ]
    optional_objects: List[Dict[str, str]] = []

    # pylint: disable=W0102
    # pylint: disable=W0613
    @typechecked
    def get_next_actions(
        required_objects: List[Dict[str, str]],
        optional_objects: List[Dict[str, str]],
        script: Script,  # pylint: disable=W0613
        script_description: Optional[Dict] = {},
    ) -> Union[Callable[[AutomatorDevice, Screen, Script], Dict], None]:
        """Looks at the required objects and optional objects and determines
        which actions to take next.
        An example of the next actions could be the following List:
        0. Select a textbox.
        1. Send some data to a textbox.
        2. Click on the/a "Next" button.

        Then the app goes to the next screen and waits a pre-determined
        amount, and optionally retries a pre-determined amount of attempts.
        """

        return actions_0

    return Screen(
        is_start=True,
        get_next_actions=get_next_actions,
        max_retries=max_retries,
        screen_nr=screen_nr,
        wait_time_sec=wait_time_sec,
        required_objects=required_objects,
        optional_objects=optional_objects,
    )


# pylint: disable=W0613
@typechecked
def actions_0(dev: AutomatorDevice, screen: Screen, script: Script) -> Dict:
    """Performs the actions in option 2 in this screen."""

    # Get the dictionary with apps that need to be torified.
    torifying_apps = script.input_data.torifying_apps

    if not isinstance(torifying_apps, Dict):
        raise TypeError(
            f"Error, torifying_apps not specified:{torifying_apps}"
        )
    for app_name, _ in torifying_apps.items():

        # Refresh the Orbot settings screen.
        dev(descriptionMatches="Refresh Apps").click()

        # Reload the screen data.
        unpacked_screen_dict: Dict = get_screen_as_dict(
            dev=dev,
            unpack=True,
            screen_dict={},
            reload=True,
        )

        # Map from normal function name, to name in UI xml for DAVx5 app.
        if app_name == "DAVx5":
            searched_name = "DAVx⁵"

        # Specify which dict element is required in the screen that contains
        # the app that needs to be torified.
        required_object: Dict[str, str] = {"@text": searched_name}

        # The dict that contains the app that needs to be torified contains 3
        # dicts, one for the icon, one for the name (on which is searched), and
        # another one for the checkbox. The parent dict of these 3 contains the
        # index of the button that needs to be clicked.
        item_dict = get_torified_item_index_dict(
            required_object, unpacked_screen_dict, {}
        )
        item_index = int(item_dict["@index"])

        # Click those buttons if they are not enabled.
        if not orbot_torifying_app_is_checked(item_dict):
            dev(index=item_index).click()

    # Refresh the screen.
    dev(descriptionMatches="Refresh Apps").click()

    # Click back.
    dev(descriptionContains="Navigate up").click()

    # Return the expected screens, using get_expected_screen_nrs.
    action_nr: int = int(inspect.stack()[0][3][8:])
    screen_nr: int = screen.screen_nr
    script_flow: nx.DiGraph = script.script_graph
    return {
        "expected_screens": get_expected_screen_nrs(
            G=script_flow, screen_nr=screen_nr, action_nr=action_nr
        )
    }



"""Starts a script to control an app."""

from typing import TYPE_CHECKING, Callable, Dict, List, Union

import networkx as nx
from typeguard import typechecked
from uiautomator import AutomatorDevice

if TYPE_CHECKING:
    # pylint: disable=W0406
    # from appcommander.Screen import Screen
    from appcommander.Script import Script
else:
    Script = object
    # Screen = object


# pylint: disable=E0102
# pylint: disable=R0902
# pylint: disable=R0903
class Screen:
    """Represents an Android app screen."""

    # pylint: disable=R0913
    # pylint: disable=W0102
    @typechecked
    def __init__(
        self,
        is_start: bool,
        get_next_actions: Callable[
            [Dict[str, str], Dict[str, str], Dict[str, str]],
            Union[Callable, None],
        ],
        max_retries: int,
        required_objects: List[Dict[str, str]],
        screen_nr: int,
        wait_time_sec: float,
        optional_objects: List[Dict[str, str]] = [],
    ) -> None:
        self.get_next_actions: Callable[
            [Dict[str, str], Dict[str, str], Dict[str, str]],
            Union[Callable, None],
        ] = get_next_actions

        self.is_start = is_start
        self.max_retries: int = max_retries

        """Sets the required objects for this screen.

        (If these objects are not found within the screen information
        returned by the dev, the screen will not be recogniszed. If
        it is, the screen is recognised by the: is_expected_screen function.
        """
        self.required_objects: List[Dict[str, str]] = required_objects

        """Some buttons/obtjects in the screen may appear depending on
        parameters that are not predictable in advance, e.g. whether some
        server responds or not.

        Yet some actions may depend on the presence and/or value of
        these objects. Hence they should be stored here.
        """
        self.optional_objects: Union[
            None, List[Dict[str, str]]
        ] = optional_objects

        # eloping typed dict.
        self.wait_time_sec: float = wait_time_sec
        self.screen_nr = screen_nr
        self.screen_dict: Dict = {}


@typechecked
def get_next_screen(
    current_screen_nr: str,
    script_graph: nx.DiGraph,
    actions: Union[Callable[[AutomatorDevice, Screen, Script], Dict], None],
) -> bool:
    """Gets the next expected screen."""

    neighbour_edges = []
    neighbour_names = []
    edge_actions = []

    for neighbour_name in nx.all_neighbors(script_graph, current_screen_nr):
        # Get neighbours.
        neighbour_names.append(neighbour_name)

        # Get edges twoards neighbours. (Outgoing edges).
        neighbour_edges.append([current_screen_nr, neighbour_name])

        # Get all action lists in all those outgoing edges.
        edge_actions.append(script_graph.edges[neighbour_edges[-1]].actions)

    # Verify the sought action list is in that list of lists, only once,
    # otherwise raise error.
    if actions not in edge_actions:
        raise LookupError(
            f"Error, the expected action list:{actions} was not in the "
            + f"available actions:{edge_actions}"
        )

    for edge in edge_actions:
        if edge.actions == actions:
            # Return the node of the next screen of the edge that contains the
            # sought action list.
            return edge[1]
    raise NotImplementedError(
        "This error should never be reached, look at code."
    )
